BLACK_LIST_DOMAINS = {
    "cdnjs.cloudflare.com",
    "ajax.googleapis.com",
    "maxcdn.bootstrapcdn.com",
    "cdn.statuspage.io",
    "static.getclicky.com"
}

BLACK_LIST_TLD = {
    "cloudfront.net",
    "etclicky.com",
    "eurolandir.com",
    "trafficmanager.net",
    "googleapis.com",
    "gstatic.com",
    "facebook.com",
    "elb.amazonaws.com",
    "amakaiedge.net",
    "googlehosted.com",
    "edgekey.net",
    "s.w.org",
    "slideshare.net",
    "linkedin.com",
    "omniture.com",
    "2O7.net",
    "twitter.com",
    "facebook.com",
    "google.com",
    "youtube.com",
    "googletagmanager.com",
    "wordpress.org",
    "wordpress.com",
    "ghs.googlehosted.com",
    "itunes.apple.com",
    "instagram.com",
    "spotify.com",
    "akadns.net",
    "fontawesome.com",
    "snapchat.com",
    "vimeo.com",
    "gmpg.org",
    "office365.com",
    "office.com",
    "jquery.org",
    "adobe.com",
    "adobelogin.com",
    "stackoverflow.com"
    "jqueryui.com",
    "jquerymobile.com",
    "adform.net",
    "packtpub.com",
    "mozilla.org",
    "w3c.org",
    "stackexchange.com",
    "stackoverflow.com",
    "serverfault.com",
    "superuser.com"
}

BLACK_LIST_PREFISES = {
    "cdn"
}



import asyncio
import xml.etree.ElementTree as et

from typing import List
from dataclasses import dataclass

import aiohttp
import filetype

FILE_TYPES = [
    filetype.is_audio,
    filetype.is_font,
    filetype.is_image,
    filetype.is_video
]


@dataclass
class S3Bucket:
    domain: str
    bucket_name: str
    objects: List[str]


def get_redirection(text: str or bytes) -> str:
    """Parse S3 XML redirection """
    root = et.fromstring(text)

    return root.find("Endpoint").text


async def download_content_and_index(
        url: str,
        bucket_name: str,
        sem: asyncio.Semaphore,
        fulltext_add_fn
):
    async with sem:
        async with aiohttp.ClientSession() as session:
            async with session.get(f"{bucket_name}/{url}") as response:
                content = await response.read()

            # Only storage non-binary files
            if not any(f(content) for f in FILE_TYPES):
                await fulltext_add_fn(bucket_name, url, content)


async def download_s3_objects(bucket: S3Bucket, fulltext_add_fn):
    sem = asyncio.Semaphore(20)

    await asyncio.gather(*[
        download_content_and_index(
            url,
            bucket.bucket_name,
            sem,
            fulltext_add_fn
        ) for url in bucket.objects
    ])


def parse_result(content: str or bytes) -> List[str]:
    """Parse S3 XML Content """
    root = et.fromstring(content)

    contents = []
    # Search contents in the bucket
    for obj in root.findall(
            "{http://s3.amazonaws.com/doc/2006-03-01/}Contents"):
        contents.append(
            obj.find("{http://s3.amazonaws.com/doc/2006-03-01/}Key").text
        )

    return contents


__all__ = ("parse_result", "S3Bucket", "get_redirection",
           "download_s3_objects")



import json
import asyncio
import argparse

from typing import List, Tuple
from functools import partial

import aiofiles

from .s3 import download_s3_objects, S3Bucket
from .redis import redis_create_connection, redis_add_document

STOP_KEYWORD = "########STOP########"


async def on_result_print_results(cli_args, bucket):

    try:
        print(f"[[[FOUND]]]] '{bucket.domain}' - Found {len(bucket.objects)}"
              f"public objects")

        if cli_args.debug:
            for obj in bucket.objects:
                print(f"        -> {bucket.domain}/{obj}")

    finally:
        if not bucket.objects:
            print(f"    *> '{bucket.domain}' - Found {len(bucket.objects)}")


async def on_result_save_streaming_results(cli_args, bucket):
    async with aiofiles.open(cli_args.result_file, mode='a+') as f:
        await f.write(f"{json.dumps(bucket.__dict__)}\n")


async def on_domain_save_new_domains(cli_args,
                                     domain: str,
                                     file_name: str,
                                     initial_domains: List[str]):

    if initial_domains and domain in initial_domains:
        return

    async with aiofiles.open(file_name, mode='a+') as f:
        await f.write(f"{domain}\n")


async def on_results_add_to_redis(
        cli_args: argparse.Namespace,
        bucket: S3Bucket):

    print(f"    >> Indexing content for '{bucket.domain}'")
    redis_con = await redis_create_connection(cli_args.index_server)

    fulltext_add_fn = partial(redis_add_document, redis_con)

    await download_s3_objects(bucket, fulltext_add_fn)



async def on_domain_event(cli_args,
                          domain_queue: asyncio.Queue,
                          initial_domains: List or None,
                          consumers: List[Tuple]):
    while True:

        domain = await domain_queue.get()

        if domain == STOP_KEYWORD:
            break

        for fn, filename in consumers:
            await fn(cli_args, domain, filename, initial_domains)


async def on_result_event(cli_args,
                          results_queue: asyncio.Queue,
                          consumers: List):

    while True:

        bucket = await results_queue.get()

        if bucket == STOP_KEYWORD:
            return

        for c in consumers:
            await c(cli_args, bucket)


__all__ = ("on_result_event", "on_domain_event", "on_domain_save_new_domains",
           "on_result_print_results", "on_result_save_streaming_results",
           "on_results_add_to_redis", "STOP_KEYWORD")



import asyncio
import argparse

from urllib.parse import urlparse

import aiohttp
import aiohttp_proxy

from lxml import etree
from async_dns.core import types
from colorama import Fore, Back, Style
from async_dns.resolver import ProxyResolver
from aiohttp_proxy import ProxyConnector, ProxyType

from . import valid_domain_or_link
from .s3 import get_redirection, parse_result, S3Bucket

PS = f"{Fore.YELLOW}SKIP{Style.RESET_ALL}"
PB = f"{Fore.GREEN}BUCKET{Style.RESET_ALL}"
PBE = f"{Fore.RED}BUCKET-ERROR{Style.RESET_ALL}"
PC= f"{Fore.MAGENTA}CRAWLER-S3-LINK{Style.RESET_ALL}"
PC3= f"{Fore.MAGENTA}CRAWLER-S3{Style.RESET_ALL}"
PCE= f"{Fore.RED}CRAWLER-ERROR{Style.RESET_ALL}"
PD= f"{Fore.BLUE}DNS{Style.RESET_ALL}"
PDE= f"{Fore.RED}DNS-ERROR{Style.RESET_ALL}"

def build_tor_connector(cli_args: argparse.Namespace) \
        -> ProxyConnector or None:

    if cli_args.tor:
        return ProxyConnector(
            proxy_type=ProxyType.SOCKS5,
            host='127.0.0.1',
            port=9050,
            verify_ssl=False
        )
    else:
        return None

async def check_tor_connection(cli_args) -> bool:

    try:
        async with aiohttp.ClientSession(connector=build_tor_connector(
                cli_args),
                timeout=aiohttp.ClientTimeout(total=cli_args.http_timeout)
        ) as session:
            async with session.get("https://www.google.com") as response:
                return True
    except aiohttp_proxy.errors.SocksConnectionError as e:
        return False


class BucketRedirectException(Exception):

    def __init__(self, redirection):
        self.redirection = redirection


async def get_bucket_info(cli_args, domain, bucket_name: str):
    async with aiohttp.ClientSession(connector=build_tor_connector(
            cli_args),
            timeout=aiohttp.ClientTimeout(total=cli_args.http_timeout)
    ) as session:

        if not bucket_name.startswith("http"):
            bucket_name = f"http://{bucket_name}"

        async with session.get(bucket_name) as response:

            if str(response.status).startswith("2"):
                content = await response.text()

                if objects := parse_result(content):
                    print(f"[{PB}] Found "
                          f"'{len(objects)}' objects at "
                          f"bucket '{bucket_name}'")

                    yield S3Bucket(
                        domain=domain,
                        bucket_name=bucket_name,
                        objects=[path for path in objects]
                    )

            elif response.status == 301:
                redirection_url = get_redirection(await response.read())

                raise BucketRedirectException(redirection_url)


async def get_links(cli_args: argparse.Namespace,
                    domain: str,
                    recursion_level: int,
                    input_queue: asyncio.Queue,
                    results_queue: asyncio.Queue):
    debug = cli_args.debug
    quiet = cli_args.quiet

    # Get links
    # found_domains= {"http": set(), "https": set()}
    found_domains= {}

    found = []

    for scheme in ("http", "https"):
        try:
            async with aiohttp.ClientSession(connector=build_tor_connector(
                    cli_args),
                    timeout=aiohttp.ClientTimeout(total=cli_args.http_timeout)
            ) as session:

                async with session.get(f"{scheme}://{domain}",
                                       verify_ssl=False) as response:
                    content = await response.text()

                    header_content_type = response.headers.get("Content-Type", "")

                    if "xml" in header_content_type:
                        content_type = "xml"
                    elif "html" in header_content_type:
                        content_type = "html"
                    else:
                        continue

                    if hasattr(content, "encode"):
                        content = content.encode("UTF-8")

                    found_domains[scheme] = (
                        content,
                        content_type,
                        response.status
                    )

        except asyncio.exceptions.TimeoutError as e:
            if debug:
                print(f"[{PCE}] Error in 'get_links'. Timeout Error "
                      f"for '{scheme}://{domain}'")
        except Exception as e:
            if debug:
                print(f"[{PCE}] Error in 'get_links': {str(e)}")
            continue

    #
    # Analyze found links
    #
    already_added_domains = set()

    for scheme, (content, content_type, status_code) in found_domains.items():

        origin = f"{scheme}://{domain}"

        if content_type == "html":

            if not content:
                continue

            tree = etree.HTML(content)

            try:
                for link in list(tree.xpath(".//@href") + tree.xpath(".//@src")):
                    link_domain = urlparse(link).netloc

                    if not link_domain:
                        continue

                    if link_domain in already_added_domains:
                        continue
                    else:
                        already_added_domains.add(link_domain)

                    if message := valid_domain_or_link(link_domain):
                        print(message)
                        continue

                    if not quiet:
                        if "s3." in link:
                            print(f"[{PC}] "
                                  f"Possible s3 bucket found. "
                                  f"'{origin}' -> "
                                  f"'{link_domain}'", flush=True)
                        else:
                            print(f"[{PC3}] Adding "
                                  f"domain to proposal. "
                                  f"{origin} -> '{link_domain}'", flush=True)

                    await input_queue.put((link_domain, recursion_level - 1))
            except AttributeError as e:
                print(f"[{PCE}] Error in parsing response from '{domain}'"
                      f": {str(e)}")
                continue

        if content_type == "xml":

            try:
                if str(status_code).startswith("2"):
                    if objects := parse_result(content):
                        print(f"[{PB}] Found "
                              f"'{len(objects)}' objects at "
                              f"bucket '{origin}'")

                        await results_queue.put(S3Bucket(
                            domain=domain,
                            bucket_name=origin,
                            objects=[path for path in objects]
                        ))

                elif response.status == 301:
                    redirection_url = get_redirection(await response.read())

                    await input_queue.put(
                        (redirection_url, recursion_level - 1)
                    )
            except Exception as e:
                # Parser error
                continue

async def get_dns_info(cli_args: argparse.Namespace,
                       domain: str,
                       recursion_level: int,
                       input_queue: asyncio.Queue):

    debug = cli_args.debug

    if cli_args.dns_resolver:
        dns_servers =[(None, cli_args.dns_resolver.split(","))]
    else:
        dns_servers = None
    try:
        resolver = ProxyResolver(proxies=dns_servers)
    except Exception as e:
        pass

    for _ in range(3):
        try:
            cname_response = await resolver.query(domain, types.CNAME)
        except Exception as e:
            if debug:
                print(f"[{PDE}] Error in 'get_dns_info': : {str(e)}")
            return

        try:
            for resp in cname_response.an:
                if resp.data and resp.qtype==types.CNAME:
                    print(f"[{PD}] Found new CNAME. '{domain}' -> "
                          f"'{resp.data}'", flush=True)

                    if message := valid_domain_or_link(resp.data):
                        print(message)
                        continue

                    await input_queue.put((resp.data, recursion_level - 1))

            break

        except AttributeError:
            await asyncio.sleep(1)


async def get_s3(cli_args: argparse.Namespace,
                 domain: str,
                 recursion_level: int,
                 input_queue: asyncio.Queue,
                 results_queue: asyncio.Queue):

    debug = cli_args.debug
    quiet = cli_args.quiet

    try:

        if domain.endswith("s3.amazonaws.com"):
            bucket_name = domain
        elif "s3" in domain:
            _s = domain.find("s3")  # Another S3 provider
            provider = domain[_s:]
            domain = domain[:_s - 1]
            bucket_name = f"http://{provider}/{domain}"
        else:
            bucket_name = "https://s3.amazonaws.com/{domain}".format(
                domain=domain
            )

            try:
                async for bucket in get_bucket_info(
                        cli_args,
                        domain,
                        bucket_name
                ):
                    await results_queue.put(bucket)
            except BucketRedirectException as red:
                if quiet:
                    print(
                        f"[{PB}] Found a redirection for bucket "
                        f"'{domain}' -> {red.redirection}",
                        flush=True)

                await input_queue.put((red.redirection, recursion_level - 1))


    except asyncio.exceptions.TimeoutError as e:
        if debug:
            print(f"[{PBE}] Error in 'get_s3'. Timeout Error "
                  f"for '{bucket_name}'")
    except Exception as e:
        if debug:
            print(f"[{PBE}] Error in 'get_s3': {str(e)} ")


__all__ = ("get_s3", "get_dns_info", "get_links", "check_tor_connection")



from colorama import Fore, Back, Style

from .black_list import *


def valid_domain_or_link(domain_or_link: str) -> None or str:
    colored_prefix = f"{Fore.YELLOW}SKIP{Style.RESET_ALL}"
    if any(domain_or_link.endswith(d) for d in BLACK_LIST_TLD):
        return f"[{colored_prefix}] domain '{domain_or_link}' is in blacklist"

    if any(domain_or_link.startswith(d) for d in BLACK_LIST_PREFISES):
        return f"[{colored_prefix}] domain '{domain_or_link}' has a prefix " \
               f"blacklisted"

    if domain_or_link in BLACK_LIST_DOMAINS:
        return f"[{colored_prefix}] domain '{domain_or_link}' is in blacklist"

    return None


__all__ = ("valid_domain_or_link",)



import re
import os
import asyncio
import argparse
import platform

from typing import Set

import aiofiles
import pkg_resources

from colorama import Fore, Style
from watchgod import awatch

from festin import *
from festin.events import *

SK = f"{Fore.YELLOW}SKIP{Style.RESET_ALL}"
SKR = f"{Fore.CYAN}SKIP-RECURSION{Style.RESET_ALL}"

async def analyze(cli_args: argparse.Namespace,
                  domain: str,
                  recursion_level: int,
                  results_queue: asyncio.Queue,
                  sem: asyncio.Semaphore,
                  input_domains_queue: asyncio.Queue):

    tasks = []
    try:
        #
        # Getting info from AWS
        #
        t1 = asyncio.create_task(get_s3(
            cli_args,
            domain,
            recursion_level,
            input_domains_queue,
            results_queue)
        )

        tasks.append(t1)

    except Exception as e:
        print(e)

    try:
        #
        # Get web links?
        #
        if not cli_args.no_links:
            t2 = asyncio.create_task(get_links(
                cli_args,
                domain,
                recursion_level,
                input_domains_queue,
                results_queue)
            )

            tasks.append(t2)

    except Exception as e:
        print(e)

    try:
        #
        # Get cnames
        #
        # if cli_args.dns:
        if not cli_args.no_dnsdiscover:
            t3 = asyncio.create_task(get_dns_info(
                cli_args,
                domain,
                recursion_level,
                input_domains_queue)
            )

            tasks.append(t3)

    except Exception as e:
        print(e)

    try:
        await asyncio.gather(*tasks)
    finally:
        sem.release()
        input_domains_queue.task_done()


async def analyze_domains(cli_args: argparse.Namespace,
                          processed_domains: Set[str],
                          results_queue: asyncio.Queue,
                          input_queue_domains: asyncio.Queue,
                          discovered_domains: asyncio.Queue,
                          raw_discovered_domains: asyncio.Queue):

    concurrency = cli_args.concurrency
    domain_regex = cli_args.domain_regex

    tasks = []

    sem = asyncio.Semaphore(value=concurrency)

    while True:

        if cli_args.watch:
            domain, recursion_level = await input_queue_domains.get()
        else:
            try:
                #
                # When we execute one shot, we need a way to stop the loop ->
                # a timeout each 5 seconds
                #
                domain, recursion_level = await asyncio.wait_for(
                    input_queue_domains.get(),
                    5
                )
            except asyncio.exceptions.TimeoutError:
                if all(t.done() for t in tasks) \
                        and input_queue_domains.empty():
                    return
                else:
                    continue

        if recursion_level < 0:
            print(f"[{SKR}] Maximum recursion level reached. Omitting "
                  f"'{domain}'")
            continue

        if hasattr(domain, "decode"):
            domain = domain.decode("UTF-8")

        await raw_discovered_domains.put(domain)

        if message := valid_domain_or_link(domain):
            print(message)

        if not domain or domain in processed_domains:
            print(f"[{SK}] domain '{domain}' already processed")
            continue

        processed_domains.add(domain)

        await discovered_domains.put(domain)

        if domain_regex:
            if not domain_regex.search(domain):
                continue

        tasks.append(
            asyncio.create_task(analyze(
                cli_args,
                domain,
                recursion_level,
                results_queue,
                sem,
                input_queue_domains
            ))
        )

        await sem.acquire()

    await asyncio.gather(*tasks)


async def run(cli_args: argparse.Namespace, init_domains: list):

    async def watch_new_domains():

        print("[*] Watching for new domains")
        async for _ in awatch(cli_args.file_domains):
            async with aiofiles.open(cli_args.file_domains, mode='r') as f:
                file_content = await f.read()

                clean_content_file = set(file_content.splitlines())

                # Select only new domains
                new_domains = clean_content_file.difference(domains_processed)

                if not new_domains:
                    print(f"[DOMAIN>>>>] Added new domain to "
                          f"'{cli_args.file_domains}' but there're already "
                          f"in file. So skipping")
                # Append new domains to processed domains and to the queue
                # domains_processed.update(new_domains)
                for d in new_domains:
                    if not d:
                        continue

                    if not quiet:
                        print(f"[DOMAIN>>>>] Added for processing: '{d}'")

                    await input_domain_queue.put(
                        (d, cli_args.http_max_recursion)
                    )

    quiet = cli_args.quiet
    domains_processed = set()
    input_domain_queue = asyncio.Queue()
    results_queue = asyncio.Queue()
    filtered_discovered_domains = asyncio.Queue()
    raw_discovered_domains = asyncio.Queue()

    #
    # Check proxy connection
    #
    if cli_args.tor:
        if not await check_tor_connection(cli_args):
            print('[!] Can\'t stablish connection to TOR')
            exit(1)

    #
    # Populate initial domains
    #
    for d in init_domains:
        input_domain_queue.put_nowait(
            (d, cli_args.http_max_recursion)
        )

    #
    # On results events
    #
    on_results_tasks = []

    if cli_args.index:
        on_results_tasks.append(on_results_add_to_redis)

    if not cli_args.result_file:
        cli_args.result_file = "results.fetin"
    on_results_tasks.append(on_result_save_streaming_results)

    if not cli_args.no_print or not cli_args.quiet:
        on_results_tasks.append(on_result_print_results)

    #
    # On domain events
    #
    on_domain_filtered_tasks = []
    on_domain_raw_domains_tasks = []

    if cli_args.discovered_domains:
        on_domain_filtered_tasks.append((
            on_domain_save_new_domains,
            cli_args.discovered_domains
        ))
    if cli_args.raw_discovered_domains:
        on_domain_raw_domains_tasks.append((
            on_domain_save_new_domains,
            cli_args.raw_discovered_domains
        ))

    #
    # Launch services
    #
    wait_tasks = []

    wait_tasks.append(asyncio.create_task(
        on_result_event(cli_args, results_queue, on_results_tasks)
    ))
    wait_tasks.append(asyncio.create_task(
        on_domain_event(cli_args,
                        filtered_discovered_domains,
                        init_domains,
                        on_domain_filtered_tasks)
    ))
    wait_tasks.append(asyncio.create_task(
        on_domain_event(cli_args,
                        raw_discovered_domains,
                        None,
                        on_domain_raw_domains_tasks)
    ))

    # Launch watcher
    if cli_args.watch:
        wait_tasks.append(
            asyncio.create_task(watch_new_domains())
        )

    #
    # Run initial discover
    #
    try:
        await analyze_domains(
            cli_args,
            domains_processed,
            results_queue,
            input_domain_queue,
            filtered_discovered_domains,
            raw_discovered_domains
        )
    finally:
        if not cli_args.watch:
            await results_queue.put(STOP_KEYWORD)
            await filtered_discovered_domains.put(STOP_KEYWORD)
            await raw_discovered_domains.put(STOP_KEYWORD)

    #
    # Wait for all tasks finish
    #
    await asyncio.wait(wait_tasks)



def main():

    #
    # Check python version
    #
    if platform.python_version_tuple() < ("3", "8"):
        print("\n[!] Python 3.8 or above is required\n")
        print("If you don't want to install Python 3.8. "
              "Try with Docker:\n")
        print("   $ docker run --rm cr0hn/festin -h")
        exit(1)

    parser = argparse.ArgumentParser(
        description='Festin - the '
                    'powered S3 bucket finder and content discover'
    )

    parser.add_argument("domains", nargs="*")
    parser.add_argument("--version",
                        action="store_true",
                        default=False,
                        help="show version")
    parser.add_argument("-f", "--file-domains",
                        default=None,
                        help="file with domains")
    parser.add_argument("-w", "--watch",
                        action="store_true",
                        default=False,
                        help="watch for new domains in file domains '-f' "
                             "option")
    parser.add_argument("-c", "--concurrency",
                        default=5,
                        type=int,
                        help="max concurrency")

    group_http = parser.add_argument_group('HTTP Probes')
    group_http.add_argument("--no-links",
                            action="store_false",
                            default=False,
                            help="extract web site links")
    group_http.add_argument("-T", "--http-timeout",
                            type=int,
                            default=5,
                            help="set timeout for http connections")
    group_http.add_argument("-M", "--http-max-recursion",
                            type=int,
                            default=3,
                            help="maximum recursison when follow links")
    group_http.add_argument("-dr", "--domain-regex",
                           default=None,
                           help="only follow domains that matches this regex")

    group_results = parser.add_argument_group('Results')
    group_results.add_argument("-rr", "--result-file",
                               default=None,
                               help="results file")
    group_results.add_argument("-rd", "--discovered-domains",
                               default=None,
                               help="file name for storing new discovered "
                                    "after apply filters")
    group_results.add_argument("-ra", "--raw-discovered-domains",
                               default=None,
                               help="file name for storing any domain without "
                                    "filters")

    group_conn = parser.add_argument_group('Connectivity')
    group_conn.add_argument("--tor",
                            default=None,
                            action="store_true",
                            help="Use Tor as proxy")

    group_display = parser.add_argument_group('Display options')
    group_display.add_argument("--debug",
                               default=False,
                               action="store_true",
                               help="enable debug mode")
    group_display.add_argument("--no-print",
                               default=False,
                               action="store_true",
                               help="doesn't print results in screen")
    group_display.add_argument("-q", "--quiet",
                               default=False,
                               action="store_true",
                               help="Use quiet mode")

    group_redis = parser.add_argument_group('Redis Search')
    group_redis.add_argument("--index",
                             default=None,
                             action="store_true",
                             help="Download and index documents into Redis")
    group_redis.add_argument("--index-server",
                             default="redis://localhost:6379",
                             help="Redis Search Server"
                                  "Default: redis://localhost:6379")

    group_dns = parser.add_argument_group('DNS options')
    group_dns.add_argument("-dn", "--no-dnsdiscover",
                           action="store_false",
                           default=False,
                           help="not follow dns cnames")
    group_dns.add_argument("-ds", "--dns-resolver",
                           default=None,
                           help="comma separated custom domain name servers")

    parsed = parser.parse_args()

    if not parsed.quiet:
        print(LOGO)

    if parsed.version:
        print(f"version: {pkg_resources.get_distribution('festin').version}")
        print()
        exit()

    if not parsed.domain_regex:
        print()
        print("#" * 50)
        print("#                                                #")
        print("#   IT'S VERY IMPORTANT TO CONFIGURE A DOMAIN    #")
        print("#   REGEX (Option '-dr'). OTHERWISE CRAWLER      #")
        print("#   WILL FOLLOW ANY LINK NO MATTER WHERE THEY    #")
        print("#   POINT TO                                     #")
        print("#                                                #")
        print("#" * 50)
        print()

    domains = []
    if parsed.domains:
        domains.extend(parsed.domains)

    if parsed.file_domains:
        print(f"[*] Loading '{parsed.file_domains}' file")
        with open(parsed.file_domains, "r") as f:
            domains.extend(f.read().splitlines())

    # Remove duplicates
    domains = list(set(domains))

    if not domains:
        print("[!] You must provide at least one domain")
        exit(1)

    if parsed.domain_regex:
        parsed.domain_regex = re.compile(parsed.domain_regex)

    if parsed.watch:
        if not parsed.file_domains:
            print("[!] For running in 'Watch' mode you must set a domains "
                  "file ('-f' option)")
            exit(1)

    if not parsed.quiet:
        print("[*] Starting FestIN")

        try:
            asyncio.run(run(parsed, domains))
        except KeyboardInterrupt:
            print("[*] Stopping Festin")


if __name__ == '__main__':
    main()



import argparse
import hashlib
from functools import partial

import aioredis


async def redis_create_connection(connection_string: str):
    async def redis_create_index(connection):
        try:
            await connection.execute(
                "FT.CREATE",
                "s3_index",

                "SCHEMA",

                "bucket",
                "TEXT",

                "filename",
                "TEXT",

                "content",
                "TEXT",
                "WEIGHT",
                "5.0"
            )
        except Exception as e:
            if "Index already exists" not in str(e):
                raise

    redis_con = await aioredis.create_redis_pool(connection_string)

    await redis_create_index(redis_con)

    return redis_con


async def redis_add_document(connection,
                             bucket_name: str,
                             object_path: str,
                             content: bytes):
    object_id = f"{bucket_name}{object_path}".encode("utf-8")

    try:
        await connection.execute(
            "FT.ADD",
            "s3_index",
            hashlib.sha512(object_id).hexdigest(),
            "1.0",
            "FIELDS",
            "bucket",
            bucket_name,
            "filename",
            object_path,
            "content",
            content
        )
    except Exception as e:
        message = str(e)
        if "Document already exists" not in message:
            print(f"    !> Insertion error: {message}")






__all__ = ("redis_add_document", "redis_create_connection")
